import os
import re
import string
import sys
import textwrap

import common_functions as common
import nml_block

class Vehicle():
    """Container class for a vehicle."""
    feature = "FEAT_TRAINS"
    # Get cargos from cargotable. Ugly
    with open(os.path.join('src', 'nml', 'cargotable.pnml')) as cargotable:
        cargo_pat = re.compile("cargotable.*\{(.*)\}", re.DOTALL)
        cargos = re.search(cargo_pat, cargotable.read())
        cargos = set(common.listify(cargos.group(1)))

    def __init__(self, csv_dict_in):
        """
        Create a vehicle.
        """
        self.init_props(csv_dict_in)
        self.init_sprites()
        self.init_graphics()
        self.init_nml()

    def init_props(self, csv_dict_in):
        default_props = {
            "intr_y": 1920,
            "model_life": "VEHICLE_NEVER_EXPIRES",
            "vehicle_life": 255,
            "refittable_cargo_classes": 0,
            "non_refittable_cargo_classes": 0,
            "cargo_allow_refit": "",
            "cargo_disallow_refit": "",
            "loading_speed": 0,
            "cost_factor": 0,
            "running_cost_factor": 0,

            "speed": 0,
            "track_type": "RAIL",
            "power": 0,
            "default_cargo_type": 0,
            "capacity": 0,
            "weight": 0,
            "engine_class": "ENGINE_CLASS_STEAM",
            "te": 0.3,
            "length": 8,
            "visual_effect": "disable"
        }
        """Convert the props from csv to more ottd-friendly versions."""
        self.csv_data = csv_dict_in
        # Fix any missing props
        for p in default_props:
            if p not in self.csv_data or not self.csv_data[p]:
                self.csv_data[p] = default_props[p]
        # Listify
        for p in ["length", "visual_effect", "sprite_source", "sprites"]:
            self.csv_data[p] = common.listify(self.csv_data[p])
        # Replace visual effect shorthands with real ones
        for i, p in enumerate(self.csv_data["visual_effect"][:]):
            try:
                visual, offset = p.split()
            except ValueError:
                visual = p
                offset = 0
            self.csv_data["visual_effect"][i] = (
                "visual_effect_and_powered(\n"
                "{:<12}VISUAL_EFFECT_{}, {}, DISABLE_WAGON_POWER)").format(
                    '', visual.upper(), offset)
        # Add bitmask() to cargo classes
        for p in ["refittable_cargo_classes", "non_refittable_cargo_classes"]:
            if self.csv_data[p]:
                self.csv_data[p] = "bitmask({})".format(self.csv_data[p])
        # Make disallowed cargos = all cargos - allowed
        cargo_allow = set(common.listify(self.csv_data["cargo_allow_refit"]))
        cargo_disallow = self.cargos - cargo_allow
        cargo_text = textwrap.TextWrapper(
            width=79, initial_indent=' '*12, subsequent_indent=' '*12)
        for c, p in zip((cargo_allow, cargo_disallow),
                        ["cargo_allow_refit", "cargo_disallow_refit"]):
            if not list(filter(None, c)):
                self.csv_data[p] = "[]"
            else:
                self.csv_data[p] = "[\n{}\n{:<8}]".format(
                    cargo_text.fill(', '.join(c)), '')

    def init_sprites(self):
        """
        Initialize the list of pngs we are going to use, and any pngs we
        need to render from xcfs.
        """
        self.sprites = []
        # Contains the actual sprites used in spritesets etc.
        self.sprites_to_render = []
        # Contains tuples of the sprites we need to render
        for i, s in enumerate(self.csv_data["sprite_source"]):
            img_name_in = s.split(':')[0]
            png_name_out = img_name_in
            # Loop back to find missing source file names
            if not img_name_in:
                for t in reversed(self.csv_data["sprite_source"][:i]):
                    try_name = t.split(':')[0]
                    if try_name:
                        img_name_in = try_name
            if '.xcf' in img_name_in.lower():
                png_name_out = os.path.splitext(img_name_in)[0]
                layers = []
                if ';' in s:
                    layers.extend(common.listify(s.split(':')[1], ';'))
                    png_name_out += '_' + '_'.join(layers)
                png_name_out += '.png'
                self.sprites_to_render.append(
                    (png_name_out, img_name_in, ' '.join(layers)))
            self.sprites.append(png_name_out)

    def init_graphics(self):
        self.switches = []
        self.spritesets = []
        self.graphics_block = nml_block.GraphicsBlock(
            name=self.csv_data["name"])
        self.sprite_nums = {}
        # Create a "sprite nums" dict we will use later
        for i in self.csv_data["sprites"]:
            def_ = common.listify(i, ':')
            if "arti" in def_[0]:
                if not "articulated" in self.sprite_nums:
                    self.sprite_nums["articulated"] = [int(def_[1])]
                else:
                    self.sprite_nums["articulated"].append(int(def_[1]))
            if "pur" in i:
                self.sprite_nums["purchase"] = int(def_[1])

        if "articulated" in self.sprite_nums:
            for i in self.sprite_nums["articulated"]:
                self.spritesets.append(nml_block.Spriteset(
                    name=self.csv_data["name"] + "_articulated_{}".format(i),
                    source=self.sprites[i-1]))
            self.switches.append(nml_block.switch(
                name=self.csv_data["name"] + "_articulated_gfx",
                var="{} % {}".format(
                    "position_in_vehid_chain",
                    len(self.sprite_nums["articulated"])),
                stuff=[i.name for i in self.spritesets]))
            self.graphics_block.add("articulated_gfx")
        else:
            self.spritesets.append(nml_block.Spriteset(
                name=self.csv_data["name"] + "_default",
                source=self.sprites[0]))
            self.graphics_block.add("default")
        if "purchase" in self.sprite_nums:
            self.spritesets.append(nml_block.Spriteset(
                name=self.csv_data["name"] + "_purchase",
                source=self.sprites[self.sprite_nums["purchase"]-1],
                layout="purchase"))
            self.graphics_block.add("purchase")

        # Change graphics block to a string
        self.spritesets = [str(i) for i in self.spritesets]
        self.graphics_block = str(self.graphics_block)
                    
    def init_lang(self):
        pass

    def init_nml(self):
        nml_temp = """item($feature, item_$name) {
    property {
        name: string(str_name_$name);
        climates_available: ALL_CLIMATES;
        introduction_date: date($intr_y,01,01);
        model_life: $model_life;
        retire_early: 0;
        vehicle_life: $vehicle_life;
        reliability_decay: 20;
        refittable_cargo_classes: $refittable_cargo_classes;
        non_refittable_cargo_classes: $non_refittable_cargo_classes;
        cargo_allow_refit: $cargo_allow_refit;
        cargo_disallow_refit: $cargo_disallow_refit;
        loading_speed: 0;
        cost_factor: $cost_factor;
        running_cost_factor: $running_cost_factor;
        cargo_age_period: 185;

        sprite_id: SPRITE_ID_NEW_TRAIN;
        speed: $speed km/h;
        misc_flags: bitmask(TRAIN_FLAG_2CC, TRAIN_FLAG_FLIP);
        refit_cost: 0;
        track_type: $track_type;
        ai_special_flag: 0;
        power: $power hpM;
        running_cost_base: RUNNING_COST_DIESEL;
        dual_headed: 0;
        default_cargo_type: $default_cargo_type;
        cargo_capacity: $capacity;
        weight: $weight ton;
        engine_class: $engine_class;
        extra_power_per_wagon: 0;
        tractive_effort_coefficient: $te;
        air_drag_coefficient: 0;
        length: $length;
        visual_effect_and_powered: $visual_effect;
        extra_weight_per_wagon: 0;
        bitmask_vehicle_info: 0;
    }
}
"""
        self.nml = []
        # Extend spritesets
        self.nml.extend(self.spritesets)
        self.nml.extend(self.switches)
        # Append props
        self.nml.append(string.Template(nml_temp).substitute(self.csv_data,
                              feature=self.feature,
                              length=self.csv_data["length"][0],
                              visual_effect=self.csv_data["visual_effect"][0],
                              graphics_block=""))
        # Append graphics block
        self.nml.append(self.graphics_block)

    def get_gfx_deps(
            self,build='build', source=os.path.join('src', 'gfx'),
            csv='$(CSV_SRC)'):
        out = []
        for s in self.sprites_to_render:
            png_out = os.path.join(build, s[0])
            xcf_in = os.path.join(source, s[1])
            out.append("render_graphics: {}".format(png_out))
            out.append("{}: {} {}".format(png_out, xcf_in, csv))
            out.append((
                '\t$(_E) "[XCF2PNG] $@"\n'
                '\t$(_V) mkdir -p $(dir $@)\n'
                '\t$(_V) xcf2png --autocrop --background \\#FFFFFF '
                '--dissolve \\\n\t      --output {}').format(
                    ' '.join([png_out, xcf_in, s[2]])))
        return '\n'.join(out)

    def get_nml(self):
        return '\n'.join(self.nml)

    def get_lang(self):
        return 'str_name_{}: {}'.format(
            self.csv_data["name"], self.csv_data["long_name"])
        pass
